import 'package:arna/arna.dart';
import 'package:flutter/cupertino.dart' show DefaultCupertinoLocalizations;
import 'package:flutter/material.dart' show DefaultMaterialLocalizations;
import 'package:flutter/services.dart' show LogicalKeyboardKey, RawKeyDownEvent;

//TODO: Add routerConfig after 2.13

/// An application that uses Arna design.
///
/// It builds upon a [WidgetsApp] by Arna specific defaulting such as fonts and scrolling physics.
///
/// The [ArnaApp] configures the top-level [Navigator] to search for routes in the following order:
///
///  1. For the `/` route, the [home] property, if non-null, is used.
///
///  2. Otherwise, the [routes] table is used, if it has an entry for the route.
///
///  3. Otherwise, [onGenerateRoute] is called, if provided. It should return a non-null value for any _valid_ route
///     not handled by [home] and [routes].
///
///  4. Finally if all else fails [onUnknownRoute] is called.
///
/// If [home], [routes], [onGenerateRoute], and [onUnknownRoute] are all null, and [builder] is not null, then no
/// [Navigator] is created.
///
/// This widget also configures the observer of the top-level [Navigator] (if any) to perform [Hero] animations.
///
/// The [ArnaApp] widget isn't a required ancestor for other Arna widgets, but many Arna widgets could depend on the
/// [ArnaTheme] widget, which the [ArnaApp] composes.
///
/// {@tool snippet}
/// This example shows how to create a [ArnaApp] that disables the "debug" banner with a [home] route that will be
/// displayed when the app is launched.
///
/// ```dart
/// const ArnaApp(
///   home: ArnaScaffold(
///     body: Container(),
///   ),
///   debugShowCheckedModeBanner: false,
/// )
/// ```
/// {@end-tool}
///
/// {@tool snippet}
/// This example shows how to create a [ArnaApp] that uses the [routes] `Map` to define the "home" route and an "about"
/// route.
///
/// ```dart
/// ArnaApp(
///   routes: <String, WidgetBuilder>{
///     '/': (BuildContext context) {
///       return const ArnaScaffold(
///         body: Container(),
///       );
///     },
///     '/about': (BuildContext context) {
///       return const ArnaScaffold(
///         body: Container(),
///       );
///     }
///   },
/// )
/// ```
/// {@end-tool}
///
/// {@tool snippet}
/// This example shows how to create a [ArnaApp] that defines a [theme] that will be used for Arna widgets in the app.
///
/// ```dart
/// const ArnaApp(
///   theme: ArnaThemeData(
///     brightness: Brightness.dark,
///   ),
///   home: ArnaScaffold(
///     body: Container(),
///   ),
/// )
/// ```
/// {@end-tool}
///
/// See also:
///
///  * [ArnaScaffold], which provides a standard page layout.
///  * [Navigator], which is used to manage the app's stack of pages.
///  * [WidgetsApp], which defines the basic app elements but does not depend on the Arna library.
class ArnaApp extends StatefulWidget {
  /// Creates an ArnaApp.
  ///
  /// At least one of [home], [routes], [onGenerateRoute], or [builder] must be non-null. If only [routes] is given, it
  /// must include an entry for the [Navigator.defaultRouteName] (`/`), since that is the route used when the
  /// application is launched with an intent that specifies an otherwise unsupported route.
  ///
  /// This class creates an instance of [WidgetsApp].
  ///
  /// The boolean arguments, [routes], and [navigatorObservers], must not be null.
  const ArnaApp({
    Key? key,
    this.navigatorKey,
    this.home,
    this.theme,
    Map<String, WidgetBuilder> this.routes = const <String, WidgetBuilder>{},
    this.initialRoute,
    this.onGenerateRoute,
    this.onGenerateInitialRoutes,
    this.onUnknownRoute,
    List<NavigatorObserver> this.navigatorObservers = const <NavigatorObserver>[],
    this.builder,
    this.title = '',
    this.onGenerateTitle,
    this.color,
    this.locale,
    this.localizationsDelegates,
    this.localeListResolutionCallback,
    this.localeResolutionCallback,
    this.supportedLocales = const <Locale>[Locale('en', 'US')],
    this.showPerformanceOverlay = false,
    this.checkerboardRasterCacheImages = false,
    this.checkerboardOffscreenLayers = false,
    this.showSemanticsDebugger = false,
    this.debugShowCheckedModeBanner = true,
    this.shortcuts,
    this.actions,
    this.restorationScopeId,
    this.scrollBehavior,
    this.useInheritedMediaQuery = false,
  })  : routeInformationProvider = null,
        routeInformationParser = null,
        routerDelegate = null,
        backButtonDispatcher = null,
        super(key: key);

  /// Creates a [ArnaApp] that uses the [Router] instead of a [Navigator].
  const ArnaApp.router({
    Key? key,
    this.routeInformationProvider,
    required RouteInformationParser<Object> this.routeInformationParser,
    required RouterDelegate<Object> this.routerDelegate,
    this.backButtonDispatcher,
    this.theme,
    this.builder,
    this.title = '',
    this.onGenerateTitle,
    this.color,
    this.locale,
    this.localizationsDelegates,
    this.localeListResolutionCallback,
    this.localeResolutionCallback,
    this.supportedLocales = const <Locale>[Locale('en', 'US')],
    this.showPerformanceOverlay = false,
    this.checkerboardRasterCacheImages = false,
    this.checkerboardOffscreenLayers = false,
    this.showSemanticsDebugger = false,
    this.debugShowCheckedModeBanner = true,
    this.shortcuts,
    this.actions,
    this.restorationScopeId,
    this.scrollBehavior,
    this.useInheritedMediaQuery = false,
  })  : navigatorObservers = null,
        navigatorKey = null,
        onGenerateRoute = null,
        home = null,
        onGenerateInitialRoutes = null,
        onUnknownRoute = null,
        routes = null,
        initialRoute = null,
        super(key: key);

  /// {@macro flutter.widgets.widgetsApp.navigatorKey}
  final GlobalKey<NavigatorState>? navigatorKey;

  /// {@macro flutter.widgets.widgetsApp.home}
  final Widget? home;

  /// The top-level [ArnaTheme] styling.
  final ArnaThemeData? theme;

  /// The application's top-level routing table.
  ///
  /// When a named route is pushed with [Navigator.pushNamed], the route name is looked up in this map. If the name is
  /// present, the associated [widgets.WidgetBuilder] is used to construct a [ArnaPageRoute] that performs an
  /// appropriate transition, including [Hero] animations, to the new route.
  ///
  /// {@macro flutter.widgets.widgetsApp.routes}
  final Map<String, WidgetBuilder>? routes;

  /// {@macro flutter.widgets.widgetsApp.initialRoute}
  final String? initialRoute;

  /// {@macro flutter.widgets.widgetsApp.onGenerateRoute}
  final RouteFactory? onGenerateRoute;

  /// {@macro flutter.widgets.widgetsApp.onGenerateInitialRoutes}
  final InitialRouteListFactory? onGenerateInitialRoutes;

  /// {@macro flutter.widgets.widgetsApp.onUnknownRoute}
  final RouteFactory? onUnknownRoute;

  /// {@macro flutter.widgets.widgetsApp.navigatorObservers}
  final List<NavigatorObserver>? navigatorObservers;

  /// {@macro flutter.widgets.widgetsApp.routeInformationProvider}
  final RouteInformationProvider? routeInformationProvider;

  /// {@macro flutter.widgets.widgetsApp.routeInformationParser}
  final RouteInformationParser<Object>? routeInformationParser;

  /// {@macro flutter.widgets.widgetsApp.routerDelegate}
  final RouterDelegate<Object>? routerDelegate;

  /// {@macro flutter.widgets.widgetsApp.backButtonDispatcher}
  final BackButtonDispatcher? backButtonDispatcher;

  /// {@macro flutter.widgets.widgetsApp.builder}
  final TransitionBuilder? builder;

  /// {@macro flutter.widgets.widgetsApp.title}
  ///
  /// This value is passed unmodified to [WidgetsApp.title].
  final String title;

  /// {@macro flutter.widgets.widgetsApp.onGenerateTitle}
  ///
  /// This value is passed unmodified to [WidgetsApp.onGenerateTitle].
  final GenerateAppTitle? onGenerateTitle;

  /// {@macro flutter.widgets.widgetsApp.color}
  final Color? color;

  /// {@macro flutter.widgets.widgetsApp.locale}
  final Locale? locale;

  /// {@macro flutter.widgets.widgetsApp.localizationsDelegates}
  ///
  /// Internationalized apps that require translations for one of the locales listed in [GlobalMaterialLocalizations]
  /// should specify this parameter and list the [supportedLocales] that the application can handle.
  ///
  /// ```dart
  /// import 'package:flutter_localizations/flutter_localizations.dart';
  /// ArnaApp(
  ///   localizationsDelegates: [
  ///     // ... app-specific localization delegate[s] here
  ///     GlobalMaterialLocalizations.delegate,
  ///     GlobalWidgetsLocalizations.delegate,
  ///   ],
  ///   supportedLocales: [
  ///     const Locale('en', 'US'), // English
  ///     const Locale('he', 'IL'), // Hebrew
  ///     // ... other locales the app supports
  ///   ],
  ///   // ...
  /// )
  /// ```
  ///
  /// ## Adding localizations for a new locale
  ///
  /// The information that follows applies to the unusual case of an app adding translations for a language not already
  /// supported by [GlobalMaterialLocalizations].
  ///
  /// Delegates that produce [WidgetsLocalizations] and [MaterialLocalizations] are included automatically. Apps can
  /// provide their own versions of these localizations by creating implementations of
  /// [LocalizationsDelegate<WidgetsLocalizations>] or [LocalizationsDelegate<MaterialLocalizations>] whose load
  /// methods return custom versions of [WidgetsLocalizations] or [MaterialLocalizations].
  ///
  /// For example: to add support to [MaterialLocalizations] for a locale it doesn't already support, say
  /// `const Locale('foo', 'BR')`, one could just extend [DefaultMaterialLocalizations]:
  ///
  /// ```dart
  /// class FooLocalizations extends DefaultMaterialLocalizations {
  ///   FooLocalizations(Locale locale) : super(locale);
  ///   @override
  ///   String get okButtonLabel {
  ///     if (locale == const Locale('foo', 'BR'))
  ///       return 'foo';
  ///     return super.okButtonLabel;
  ///   }
  /// }
  ///
  /// ```
  ///
  /// A `FooLocalizationsDelegate` is essentially just a method that constructs a `FooLocalizations` object. We return
  /// a [SynchronousFuture] here because no asynchronous work takes place upon "loading" the localizations object.
  ///
  /// ```dart
  /// class FooLocalizationsDelegate extends LocalizationsDelegate<MaterialLocalizations> {
  ///   const FooLocalizationsDelegate();
  ///   @override
  ///   Future<FooLocalizations> load(Locale locale) {
  ///     return SynchronousFuture(FooLocalizations(locale));
  ///   }
  ///   @override
  ///   bool shouldReload(FooLocalizationsDelegate old) => false;
  /// }
  /// ```
  ///
  /// Constructing a [ArnaApp] with a `FooLocalizationsDelegate` overrides the automatically included delegate for
  /// [MaterialLocalizations] because only the first delegate of each [LocalizationsDelegate.type] is used and the
  /// automatically included delegates are added to the end of the app's [localizationsDelegates] list.
  ///
  /// ```dart
  /// ArnaApp(
  ///   localizationsDelegates: [
  ///     const FooLocalizationsDelegate(),
  ///   ],
  ///   // ...
  /// )
  /// ```
  /// See also:
  ///
  ///  * [supportedLocales], which must be specified along with [localizationsDelegates].
  ///  * [GlobalMaterialLocalizations], a [localizationsDelegates] value which provides material localizations for many
  ///    languages.
  ///  * The Flutter Internationalization Tutorial, <https://flutter.dev/tutorials/internationalization/>.
  final Iterable<LocalizationsDelegate<dynamic>>? localizationsDelegates;

  /// {@macro flutter.widgets.widgetsApp.localeListResolutionCallback}
  ///
  /// This callback is passed along to the [WidgetsApp] built by this widget.
  final LocaleListResolutionCallback? localeListResolutionCallback;

  /// {@macro flutter.widgets.LocaleResolutionCallback}
  ///
  /// This callback is passed along to the [WidgetsApp] built by this widget.
  final LocaleResolutionCallback? localeResolutionCallback;

  /// {@macro flutter.widgets.widgetsApp.supportedLocales}
  ///
  /// It is passed along unmodified to the [WidgetsApp] built by this widget.
  ///
  /// See also:
  ///
  ///  * [localizationsDelegates], which must be specified for localized applications.
  ///  * [GlobalMaterialLocalizations], a [localizationsDelegates] value which provides material localizations for many
  ///    languages.
  ///  * The Flutter Internationalization Tutorial,
  ///  * The Flutter Internationalization Tutorial, <https://flutter.dev/tutorials/internationalization/>.
  final Iterable<Locale> supportedLocales;

  /// Turns on a performance overlay.
  ///
  /// See also:
  ///
  ///  * <https://flutter.dev/debugging/#performance-overlay>
  final bool showPerformanceOverlay;

  /// Turns on checkerboarding of raster cache images.
  final bool checkerboardRasterCacheImages;

  /// Turns on checkerboarding of layers rendered to offscreen bitmaps.
  final bool checkerboardOffscreenLayers;

  /// Turns on an overlay that shows the accessibility information reported by the framework.
  final bool showSemanticsDebugger;

  /// {@macro flutter.widgets.widgetsApp.debugShowCheckedModeBanner}
  final bool debugShowCheckedModeBanner;

  /// {@macro flutter.widgets.widgetsApp.shortcuts}
  /// {@tool snippet}
  /// This example shows how to add a single shortcut for [LogicalKeyboardKey.select] to the default shortcuts without
  /// needing to add your own [Shortcuts] widget.
  ///
  /// Alternatively, you could insert a [Shortcuts] widget with just the mapping you want to add between the
  /// [WidgetsApp] and its child and get the same effect.
  ///
  /// ```dart
  /// Widget build(BuildContext context) {
  ///   return ArnaApp(
  ///     shortcuts: <ShortcutActivator, Intent>{
  ///       ... WidgetsApp.defaultShortcuts,
  ///       const SingleActivator(LogicalKeyboardKey.select): const ActivateIntent(),
  ///     },
  ///     color: const Color(0xFFFF0000),
  ///     builder: (BuildContext context, Widget? child) {
  ///       return const Placeholder();
  ///     },
  ///   );
  /// }
  /// ```
  /// {@end-tool}
  /// {@macro flutter.widgets.widgetsApp.shortcuts.seeAlso}
  final Map<ShortcutActivator, Intent>? shortcuts;

  /// {@macro flutter.widgets.widgetsApp.actions}
  /// {@tool snippet}
  /// This example shows how to add a single action handling an [ActivateAction] to the default actions without needing
  /// to add your own [Actions] widget.
  ///
  /// Alternatively, you could insert a [Actions] widget with just the mapping you want to add between the [WidgetsApp]
  /// and its child and get the same effect.
  ///
  /// ```dart
  /// Widget build(BuildContext context) {
  ///   return ArnaApp(
  ///     actions: <Type, Action<Intent>>{
  ///       ... WidgetsApp.defaultActions,
  ///       ActivateAction: CallbackAction<Intent>(
  ///         onInvoke: (Intent intent) {
  ///           // Do something here...
  ///           return null;
  ///         },
  ///       ),
  ///     },
  ///     color: const Color(0xFFFF0000),
  ///     builder: (BuildContext context, Widget? child) {
  ///       return const Placeholder();
  ///     },
  ///   );
  /// }
  /// ```
  /// {@end-tool}
  /// {@macro flutter.widgets.widgetsApp.actions.seeAlso}
  final Map<Type, Action<Intent>>? actions;

  /// {@macro flutter.widgets.widgetsApp.restorationScopeId}
  final String? restorationScopeId;

  /// {@macro flutter.widgets.scrollBehavior}
  ///
  /// When null, defaults to [ArnaScrollBehavior].
  ///
  /// See also:
  ///
  ///  * [ScrollConfiguration], which controls how [Scrollable] widgets behave in a subtree.
  final ScrollBehavior? scrollBehavior;

  /// {@macro flutter.widgets.widgetsApp.useInheritedMediaQuery}
  final bool useInheritedMediaQuery;

  @override
  State<ArnaApp> createState() => _ArnaAppState();

  /// The [HeroController] used for Arna page transitions.
  ///
  /// Used by [ArnaApp].
  static HeroController createArnaHeroController() => HeroController();
}

/// Describes how [Scrollable] widgets behave for [ArnaApp]s.
///
/// {@macro flutter.widgets.scrollBehavior}
///
/// Setting a [ArnaScrollBehavior] will result in descendant [Scrollable] widgets.
/// When executing on desktop platforms, a [ArnaScrollbar] is applied to the child.
///
/// See also:
///
///  * [ScrollBehavior], the default scrolling behavior extended by this class.
class ArnaScrollBehavior extends ScrollBehavior {
  /// Creates an ArnaScrollBehavior that adds [ArnaScrollbar]s on desktop platforms.
  const ArnaScrollBehavior();

  @override
  Widget buildScrollbar(context, Widget child, ScrollableDetails details) {
    // When modifying this function, consider modifying the implementation in the base class as well.
    switch (axisDirectionToAxis(details.direction)) {
      case Axis.horizontal:
        return child;
      case Axis.vertical:
        switch (getPlatform(context)) {
          case TargetPlatform.linux:
          case TargetPlatform.macOS:
          case TargetPlatform.windows:
            return ArnaScrollbar(controller: details.controller, child: child);
          case TargetPlatform.android:
          case TargetPlatform.fuchsia:
          case TargetPlatform.iOS:
            return child;
        }
    }
  }

  @override
  Widget buildOverscrollIndicator(BuildContext context, Widget child, ScrollableDetails details) {
    return StretchingOverscrollIndicator(
      axisDirection: details.direction,
      child: child,
    );
  }
}

/// The [State] for a [ArnaApp].
class _ArnaAppState extends State<ArnaApp> {
  late HeroController _heroController;

  bool get _usesRouter => widget.routerDelegate != null;

  @override
  void initState() {
    super.initState();
    _heroController = ArnaApp.createArnaHeroController();
  }

  // Combine the Localizations for Material with the ones contributed by the localizationsDelegates parameter, if any.
  // Only the first delegate of a particular LocalizationsDelegate.type is loaded so the localizationsDelegate
  // parameter can be used to override _MaterialLocalizationsDelegate.
  Iterable<LocalizationsDelegate<dynamic>> get _localizationsDelegates {
    return <LocalizationsDelegate<dynamic>>[
      if (widget.localizationsDelegates != null) ...widget.localizationsDelegates!,
      DefaultMaterialLocalizations.delegate,
      DefaultCupertinoLocalizations.delegate,
    ];
  }

  Widget _inspectorSelectButtonBuilder(BuildContext context, VoidCallback onPressed) {
    return ArnaButton(icon: Icons.search, onPressed: onPressed);
  }

  Widget _arnaBuilder(BuildContext context, Widget? child) {
    ArnaThemeData? theme;

    if (widget.theme != null) {
      if (widget.theme!.brightness == null) {
        theme = MediaQuery.platformBrightnessOf(context) == Brightness.dark
            ? ArnaThemeData.dark().copyWith(accentColor: widget.theme!.accentColor)
            : ArnaThemeData.light().copyWith(accentColor: widget.theme!.accentColor);
      }
    }

    theme ??= widget.theme ?? ArnaThemeData.light();

    return AnimatedArnaTheme(
      data: theme,
      child: widget.builder != null
          ? Builder(
              builder: (BuildContext context) {
                // Why are we surrounding a builder with a builder?
                //
                // The widget.builder may contain code that invokes ArnaTheme.of(), which should return the theme we
                // selected above in AnimatedArbaTheme. However, if we invoke widget.builder() directly as the child
                // of AnimatedArnaTheme then there is no Context separating them, and the widget.builder() will not
                // find the theme. Therefore, we surround widget.builder with yet another builder so that a context
                // separates them and ArnaTheme.of() correctly resolves to the theme we passed to AnimatedArnaTheme.
                return widget.builder!(context, child);
              },
            )
          : child ?? const SizedBox.shrink(),
    );
  }

  WidgetsApp _buildWidgetApp(BuildContext context) {
    final ArnaThemeData effectiveThemeData = ArnaTheme.of(context);
    final Color color = ArnaDynamicColor.resolve(widget.color ?? effectiveThemeData.accentColor, context);

    if (_usesRouter) {
      return WidgetsApp.router(
        key: GlobalObjectKey(this),
        routeInformationProvider: widget.routeInformationProvider,
        routeInformationParser: widget.routeInformationParser!,
        routerDelegate: widget.routerDelegate!,
        backButtonDispatcher: widget.backButtonDispatcher,
        builder: _arnaBuilder,
        title: widget.title,
        onGenerateTitle: widget.onGenerateTitle,
        textStyle: effectiveThemeData.textTheme.body,
        color: color,
        locale: widget.locale,
        localizationsDelegates: _localizationsDelegates,
        localeResolutionCallback: widget.localeResolutionCallback,
        localeListResolutionCallback: widget.localeListResolutionCallback,
        supportedLocales: widget.supportedLocales,
        showPerformanceOverlay: widget.showPerformanceOverlay,
        checkerboardRasterCacheImages: widget.checkerboardRasterCacheImages,
        checkerboardOffscreenLayers: widget.checkerboardOffscreenLayers,
        showSemanticsDebugger: widget.showSemanticsDebugger,
        debugShowCheckedModeBanner: widget.debugShowCheckedModeBanner,
        inspectorSelectButtonBuilder: _inspectorSelectButtonBuilder,
        shortcuts: widget.shortcuts,
        actions: widget.actions,
        restorationScopeId: widget.restorationScopeId,
        useInheritedMediaQuery: widget.useInheritedMediaQuery,
      );
    }
    return WidgetsApp(
      key: GlobalObjectKey(this),
      navigatorKey: widget.navigatorKey,
      navigatorObservers: widget.navigatorObservers!,
      pageRouteBuilder: <T>(RouteSettings settings, WidgetBuilder builder) {
        return ArnaPageRoute<T>(settings: settings, builder: builder);
      },
      home: widget.home,
      routes: widget.routes!,
      initialRoute: widget.initialRoute,
      onGenerateRoute: widget.onGenerateRoute,
      onGenerateInitialRoutes: widget.onGenerateInitialRoutes,
      onUnknownRoute: widget.onUnknownRoute,
      builder: _arnaBuilder,
      title: widget.title,
      onGenerateTitle: widget.onGenerateTitle,
      textStyle: effectiveThemeData.textTheme.body,
      color: color,
      locale: widget.locale,
      localizationsDelegates: _localizationsDelegates,
      localeResolutionCallback: widget.localeResolutionCallback,
      localeListResolutionCallback: widget.localeListResolutionCallback,
      supportedLocales: widget.supportedLocales,
      showPerformanceOverlay: widget.showPerformanceOverlay,
      checkerboardRasterCacheImages: widget.checkerboardRasterCacheImages,
      checkerboardOffscreenLayers: widget.checkerboardOffscreenLayers,
      showSemanticsDebugger: widget.showSemanticsDebugger,
      debugShowCheckedModeBanner: widget.debugShowCheckedModeBanner,
      inspectorSelectButtonBuilder: _inspectorSelectButtonBuilder,
      shortcuts: widget.shortcuts,
      actions: widget.actions,
      restorationScopeId: widget.restorationScopeId,
      useInheritedMediaQuery: widget.useInheritedMediaQuery,
    );
  }

  @override
  Widget build(BuildContext context) {
    return ScrollConfiguration(
      behavior: widget.scrollBehavior ?? const ArnaScrollBehavior(),
      child: HeroControllerScope(
        controller: _heroController,
        child: Focus(
          canRequestFocus: false,
          onKey: (FocusNode node, RawKeyEvent event) {
            if (event is! RawKeyDownEvent || event.logicalKey != LogicalKeyboardKey.escape) {
              return KeyEventResult.ignored;
            }
            return ArnaTooltip.dismissAllToolTips() ? KeyEventResult.handled : KeyEventResult.ignored;
          },
          child: _buildWidgetApp(context),
        ),
      ),
    );
  }
}
